@@ -34,6 +34,7 @@ syslog_tag_mysqld=mysqld
 syslog_tag_mysqld_safe=mysqld_safe
 
 trap '' 1 2 3 15			# we shouldn't let anyone kill us
+trap '' 13                              # not even SIGPIPE
 
 # MySQL-specific environment variable. First off, it's not really a umask,
 # it's the desired mode. Second, it follows umask(2), not umask(3) in that
@@ -231,24 +232,26 @@ wsrep_recover_position() {
     return $ret
   fi
 
-  local wr_logfile=$(mktemp)
   local euid=$(id -u)
 
+  local wr_logfile=$(mktemp $DATADIR/wsrep_recovery.XXXXXX)
+
   [ "$euid" = "0" ] && chown $user $wr_logfile
   chmod 600 $wr_logfile
 
-  log_notice "WSREP: Running position recovery with --log_error=$wr_logfile \
-                                --pid-file="$DATADIR/`@HOSTNAME@`-recover.pid""
+  local wr_pidfile="$DATADIR/"`@HOSTNAME@`"-recover.pid"
 
-  eval_log_error "$mysqld_cmd --log_error=$wr_logfile --wsrep-recover \
-                            --pid-file="$DATADIR/`@HOSTNAME@`-recover.pid""
+  local wr_options="--log_error='$wr_logfile' --pid-file='$wr_pidfile'"
+
+  log_notice "WSREP: Running position recovery with $wr_options"
+
+  eval_log_error "$mysqld_cmd --wsrep_recover $wr_options"
 
   local rp="$(grep 'WSREP: Recovered position:' $wr_logfile)"
   if [ -z "$rp" ]; then
     local skipped="$(grep WSREP $wr_logfile | grep 'skipping position recovery')"
     if [ -z "$skipped" ]; then
-      log_error "WSREP: Failed to recover position: 
-'`cat $wr_logfile`'"
+      log_error "WSREP: Failed to recover position: " `cat $wr_logfile`;
       ret=1
     else
       log_notice "WSREP: Position recovery skipped"
@@ -260,7 +263,7 @@ wsrep_recover_position() {
     wsrep_start_position_opt="--wsrep_start_position=$start_pos"
   fi
 
-  [ $ret -eq 0 ] && rm $wr_logfile
+  rm $wr_logfile
 
   return $ret
 }
@@ -329,6 +332,7 @@ parse_arguments() {
         then
           wsrep_restart=1
         fi
+        append_arg_to_args "$arg"
         ;;
       --help) usage ;;
 
@@ -597,7 +601,6 @@ fi
 parse_arguments `$print_defaults $defaults --loose-verbose mysqld_safe safe_mysqld`
 parse_arguments PICK-ARGS-FROM-ARGV "$@"
 
-
 #
 # Try to find the plugin directory
 #
@@ -621,7 +624,6 @@ else
 fi
 plugin_dir="${plugin_dir}${PLUGIN_VARIANT}"
 
-
 # Determine what logging facility to use
 
 # Ensure that 'logger' exists, if it's requested
@@ -645,6 +647,7 @@ then
 
     # mysqld does not add ".err" to "--log-error=foo."; it considers a
     # trailing "." as an extension
+    
     if expr "$err_log" : '.*\.[^/]*$' > /dev/null
     then
         :
@@ -958,14 +961,14 @@ do
   rm -f $safe_mysql_unix_port "$pid_file"	# Some extra safety
 
   start_time=`date +%M%S`
-  
+
   # This file is checked for empty directory because 
   # a) Not having this file means the wsrep-start-position is useless - lp:1112724
   # b) Otherwise I have to check if directory is empty sans a few files like 
   # error log and others, #a is simpler.
   if [ -e $grastate_loc ];then
-    # this sets wsrep_start_position_opt
-    wsrep_recover_position "$cmd"
+  # this sets wsrep_start_position_opt
+  wsrep_recover_position "$cmd"
   else 
     log_notice "Skipping wsrep-recover for empty datadir: ${DATADIR}"
     log_notice "Assigning 00000000-0000-0000-0000-000000000000:-1 to wsrep_start_position"
