@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2007, 2012, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2007, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -16,14 +16,11 @@
 */
 
 /*
-  Functions to autenticate and handle reqests for a connection
+  Functions to authenticate and handle requests for a connection
 */
 
 #include "my_global.h"
 #include "sql_priv.h"
-#ifndef __WIN__
-#include <netdb.h>        // getservbyname, servent
-#endif
 #include "sql_audit.h"
 #include "sql_connect.h"
 #include "my_global.h"
@@ -38,6 +35,11 @@
 #include "sql_acl.h"  // acl_getroot, NO_ACCESS, SUPER_ACL
 #include "sql_callback.h"
 
+#include <algorithm>
+
+using std::min;
+using std::max;
+
 #if defined(HAVE_OPENSSL) && !defined(EMBEDDED_LIBRARY)
 /*
   Without SSL the handshake consists of one packet. This packet
@@ -59,8 +61,10 @@
 #include "wsrep_mysqld.h"
 #endif
 
+#ifndef EMBEDDED_LIBRARY
 // Increments connection count for user.
 static int increment_connection_count(THD* thd, bool use_lock);
+#endif
 
 // Uses the THD to update the global stats by user name and client IP
 void update_global_user_stats(THD* thd, bool create_user, time_t now);
@@ -100,12 +104,12 @@ int get_or_create_user_conn(THD *thd, const char *user,
   temp_len= (strmov(strmov(temp_user, user)+1, host) - temp_user)+1;
   mysql_mutex_lock(&LOCK_user_conn);
   if (!(uc = (struct  user_conn *) my_hash_search(&hash_user_connections,
-					       (uchar*) temp_user, temp_len)))
+                 (uchar*) temp_user, temp_len)))
   {
     /* First connection for user; Create a user connection object */
     if (!(uc= ((struct user_conn*)
-	       my_malloc(sizeof(struct user_conn) + temp_len+1,
-			 MYF(MY_WME)))))
+         my_malloc(sizeof(struct user_conn) + temp_len+1,
+       MYF(MY_WME)))))
     {
       /* MY_WME ensures an error is set in THD. */
       return_val= 1;
@@ -137,7 +141,7 @@ end:
 extern "C" uchar *get_key_user_stats(USER_STATS *user_stats, size_t *length,
                          my_bool not_used __attribute__((unused)))
 {
-  *length= strlen(user_stats->user);
+  *length= user_stats->user_len;
   return (uchar*) user_stats->user;
 }
 
@@ -190,6 +194,9 @@ void init_user_stats(USER_STATS *user_stats,
   strncpy(user_stats->user, user, sizeof(user_stats->user));
   strncpy(user_stats->priv_user, priv_user, sizeof(user_stats->priv_user));
 
+  user_stats->user_len=               strlen(user_stats->user);
+  user_stats->priv_user_len=          strlen(user_stats->priv_user);
+
   user_stats->total_connections=      total_connections;
   user_stats->total_ssl_connections=  total_ssl_connections;
   user_stats->concurrent_connections= concurrent_connections;
@@ -268,94 +275,6 @@ void init_thread_stats(THREAD_STATS *thread_stats,
   DBUG_VOID_RETURN;
 }
 
-void add_user_stats(USER_STATS *user_stats,
-                    uint total_connections,
-                    uint concurrent_connections,
-                    time_t connected_time,
-                    double busy_time,
-                    double cpu_time,
-                    ulonglong bytes_received,
-                    ulonglong bytes_sent,
-                    ulonglong binlog_bytes_written,
-                    ha_rows rows_fetched,
-                    ha_rows rows_updated,
-                    ha_rows rows_read,
-                    ulonglong select_commands,
-                    ulonglong update_commands,
-                    ulonglong other_commands,
-                    ulonglong commit_trans,
-                    ulonglong rollback_trans,
-                    ulonglong denied_connections,
-                    ulonglong lost_connections,
-                    ulonglong access_denied_errors,
-                    ulonglong empty_queries)
-{
-  user_stats->total_connections+=      total_connections;
-  user_stats->concurrent_connections+= concurrent_connections;
-  user_stats->connected_time+=         connected_time;
-  user_stats->busy_time+=              busy_time;
-  user_stats->cpu_time+=               cpu_time;
-  user_stats->bytes_received+=         bytes_received;
-  user_stats->bytes_sent+=             bytes_sent;
-  user_stats->binlog_bytes_written+=   binlog_bytes_written;
-  user_stats->rows_fetched+=           rows_fetched;
-  user_stats->rows_updated+=           rows_updated;
-  user_stats->rows_read+=              rows_read;
-  user_stats->select_commands+=        select_commands;
-  user_stats->update_commands+=        update_commands;
-  user_stats->other_commands+=         other_commands;
-  user_stats->commit_trans+=           commit_trans;
-  user_stats->rollback_trans+=         rollback_trans;
-  user_stats->denied_connections+=     denied_connections;
-  user_stats->lost_connections+=       lost_connections;
-  user_stats->access_denied_errors+=   access_denied_errors;
-  user_stats->empty_queries+=          empty_queries;
-}
-
-void add_thread_stats(THREAD_STATS *thread_stats,
-                    uint total_connections,
-                    uint concurrent_connections,
-                    time_t connected_time,
-                    double busy_time,
-                    double cpu_time,
-                    ulonglong bytes_received,
-                    ulonglong bytes_sent,
-                    ulonglong binlog_bytes_written,
-                    ha_rows rows_fetched,
-                    ha_rows rows_updated,
-                    ha_rows rows_read,
-                    ulonglong select_commands,
-                    ulonglong update_commands,
-                    ulonglong other_commands,
-                    ulonglong commit_trans,
-                    ulonglong rollback_trans,
-                    ulonglong denied_connections,
-                    ulonglong lost_connections,
-                    ulonglong access_denied_errors,
-                    ulonglong empty_queries)
-{
-  thread_stats->total_connections+=      total_connections;
-  thread_stats->concurrent_connections+= concurrent_connections;
-  thread_stats->connected_time+=         connected_time;
-  thread_stats->busy_time+=              busy_time;
-  thread_stats->cpu_time+=               cpu_time;
-  thread_stats->bytes_received+=         bytes_received;
-  thread_stats->bytes_sent+=             bytes_sent;
-  thread_stats->binlog_bytes_written+=   binlog_bytes_written;
-  thread_stats->rows_fetched+=           rows_fetched;
-  thread_stats->rows_updated+=           rows_updated;
-  thread_stats->rows_read+=              rows_read;
-  thread_stats->select_commands+=        select_commands;
-  thread_stats->update_commands+=        update_commands;
-  thread_stats->other_commands+=         other_commands;
-  thread_stats->commit_trans+=           commit_trans;
-  thread_stats->rollback_trans+=         rollback_trans;
-  thread_stats->denied_connections+=     denied_connections;
-  thread_stats->lost_connections+=       lost_connections;
-  thread_stats->access_denied_errors+=   access_denied_errors;
-  thread_stats->empty_queries+=          empty_queries;
-}
-
 void init_global_user_stats(void)
 {
   if (my_hash_init(&global_user_stats, system_charset_info, max_connections,
@@ -390,7 +309,7 @@ void init_global_thread_stats(void)
 extern "C" uchar *get_key_table_stats(TABLE_STATS *table_stats, size_t *length,
                                      my_bool not_used __attribute__((unused)))
 {
-  *length= strlen(table_stats->table);
+  *length= table_stats->table_len;
   return (uchar*) table_stats->table;
 }
 
@@ -412,7 +331,7 @@ void init_global_table_stats(void)
 extern "C" uchar *get_key_index_stats(INDEX_STATS *index_stats, size_t *length,
                                      my_bool not_used __attribute__((unused)))
 {
-  *length= strlen(index_stats->index);
+  *length= index_stats->index_len;
   return (uchar*) index_stats->index;
 }
 
@@ -476,8 +395,9 @@ static int increment_count_by_name(const char *name, const char *role_name,
                                                    (uchar*) name,
                                                    strlen(name))))
   {
-    if (acl_is_utility_user(thd->security_ctx->user, thd->security_ctx->host,
-                            thd->security_ctx->ip))
+    if (acl_is_utility_user(thd->security_ctx->user,
+                            thd->security_ctx->get_host()->ptr(),
+                            thd->security_ctx->get_ip()->ptr()))
       return 0;
 
     // First connection for this user or client
@@ -520,8 +440,9 @@ static int increment_count_by_id(my_thread_id id,
                                                        (uchar*) &id,
                                                        sizeof(my_thread_id))))
   {
-    if (acl_is_utility_user(thd->security_ctx->user, thd->security_ctx->host,
-        thd->security_ctx->ip))
+    if (acl_is_utility_user(thd->security_ctx->user,
+                            thd->security_ctx->get_host()->ptr(),
+                            thd->security_ctx->get_ip()->ptr()))
       return 0;
 
     // First connection for this user or client
@@ -565,8 +486,9 @@ static int increment_connection_count(THD* thd, bool use_lock)
   const char* client_string= get_client_host(thd);
   int return_value=          0;
 
-  if (acl_is_utility_user(thd->security_ctx->user, thd->security_ctx->host,
-      thd->security_ctx->ip))
+  if (acl_is_utility_user(thd->security_ctx->user,
+                          thd->security_ctx->get_host()->ptr(),
+                          thd->security_ctx->get_ip()->ptr()))
     return return_value;
 
   if (use_lock)
@@ -654,109 +576,104 @@ static void update_global_thread_stats_with_thread(THD* thd,
 // Updates the global stats of a user or client
 void update_global_user_stats(THD* thd, bool create_user, time_t now)
 {
-  if (opt_userstat)
+  char* user_string=         get_valid_user_string(thd->main_security_ctx.user);
+  const char* client_string= get_client_host(thd);
+
+  USER_STATS* user_stats;
+  THREAD_STATS* thread_stats;
+
+  if (acl_is_utility_user(thd->security_ctx->user,
+                          thd->security_ctx->get_host()->ptr(),
+                          thd->security_ctx->get_ip()->ptr()))
+    return;
+
+  mysql_mutex_lock(&LOCK_global_user_client_stats);
+
+  // Update by user name
+  if ((user_stats = (USER_STATS *) my_hash_search(&global_user_stats,
+                                                  (uchar *) user_string,
+                                                  strlen(user_string))))
   {
-    char* user_string=         get_valid_user_string(thd->main_security_ctx.user);
-    const char* client_string= get_client_host(thd);
-
-    USER_STATS* user_stats;
-    THREAD_STATS* thread_stats;
-
-    if (acl_is_utility_user(thd->security_ctx->user, thd->security_ctx->host,
-        thd->security_ctx->ip))
-      return;
-
-    mysql_mutex_lock(&LOCK_global_user_client_stats);
-
-    // Update by user name
-    if ((user_stats = (USER_STATS *) my_hash_search(&global_user_stats,
-                                                    (uchar *) user_string,
-                                                    strlen(user_string))))
-    {
-      // Found user.
-      update_global_user_stats_with_user(thd, user_stats, now);
-    }
-    else
-    {
-      // Create the entry
-      if (create_user)
-      {
-        increment_count_by_name(user_string, user_string,
-                                &global_user_stats, thd);
-      }
-    }
-
-    // Update by client IP
-    if ((user_stats = (USER_STATS *) my_hash_search(&global_client_stats,
-                                                    (uchar *) client_string,
-                                                    strlen(client_string))))
-    {
-      // Found by client IP
-      update_global_user_stats_with_user(thd, user_stats, now);
-    }
-    else
-    {
-      // Create the entry
-      if (create_user)
-      {
-        increment_count_by_name(client_string,
-                                user_string,
-                                &global_client_stats, thd);
-      }
-    }
-
-    if (opt_thread_statistics)
-    {
-      // Update by thread ID
-      if ((thread_stats = (THREAD_STATS *) my_hash_search(&global_thread_stats,
-                                                          (uchar *) &(thd->thread_id),
-                                                          sizeof(my_thread_id))))
-      {
-        // Found by thread ID
-        update_global_thread_stats_with_thread(thd, thread_stats, now);
-      }
-      else
-      {
-        // Create the entry
-        if (create_user)
-        {
-          increment_count_by_id(thd->thread_id,
-                                &global_thread_stats, thd);
-        }
-      }
-    }
-
-    thd->last_global_update_time = now;
-    thd->reset_diff_stats();
-
-    mysql_mutex_unlock(&LOCK_global_user_client_stats);
+    // Found user.
+    update_global_user_stats_with_user(thd, user_stats, now);
   }
   else
   {
-    thd->reset_diff_stats();
+    // Create the entry
+    if (create_user)
+    {
+      increment_count_by_name(user_string, user_string,
+                              &global_user_stats, thd);
+    }
   }
+
+  // Update by client IP
+  if ((user_stats = (USER_STATS *) my_hash_search(&global_client_stats,
+                                                  (uchar *) client_string,
+                                                  strlen(client_string))))
+  {
+    // Found by client IP
+    update_global_user_stats_with_user(thd, user_stats, now);
+  }
+  else
+  {
+    // Create the entry
+    if (create_user)
+    {
+      increment_count_by_name(client_string,
+                              user_string,
+                              &global_client_stats, thd);
+    }
+  }
+
+  if (opt_thread_statistics)
+  {
+    // Update by thread ID
+    if ((thread_stats = (THREAD_STATS *) my_hash_search(&global_thread_stats,
+                                                        (uchar *) &(thd->thread_id),
+                                                        sizeof(my_thread_id))))
+    {
+      // Found by thread ID
+      update_global_thread_stats_with_thread(thd, thread_stats, now);
+    }
+    else
+    {
+      // Create the entry
+      if (create_user)
+      {
+        increment_count_by_id(thd->thread_id,
+                              &global_thread_stats, thd);
+      }
+    }
+  }
+
+  thd->last_global_update_time = now;
+  thd->reset_diff_stats();
+
+  mysql_mutex_unlock(&LOCK_global_user_client_stats);
 }
 
 /*
   check if user has already too many connections
-  
+
   SYNOPSIS
   check_for_max_user_connections()
-  thd			Thread handle
-  uc			User connect object
+  thd     Thread handle
+  uc      User connect object
 
   NOTES
     If check fails, we decrease user connection count, which means one
     shouldn't call decrease_user_connections() after this function.
 
   RETURN
-    0	ok
-    1	error
+    0 ok
+    1 error
 */
 
 int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
 {
   int error=0;
+  Host_errors errors;
   DBUG_ENTER("check_for_max_user_connections");
 
   mysql_mutex_lock(&LOCK_user_conn);
@@ -766,6 +683,7 @@ int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
   {
     my_error(ER_TOO_MANY_USER_CONNECTIONS, MYF(0), uc->user);
     error=1;
+    errors.m_max_user_connection= 1;
     goto end;
   }
   thd->time_out_user_resource_limits();
@@ -776,6 +694,7 @@ int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
              "max_user_connections",
              (long) uc->user_resources.user_conn);
     error= 1;
+    errors.m_max_user_connection= 1;
     goto end;
   }
   if (uc->user_resources.conn_per_hour &&
@@ -785,6 +704,7 @@ int check_for_max_user_connections(THD *thd, const USER_CONN *uc)
              "max_connections_per_hour",
              (long) uc->user_resources.conn_per_hour);
     error=1;
+    errors.m_max_user_connection_per_hour= 1;
     goto end;
   }
   thd->increment_con_per_hour_counter();
@@ -802,6 +722,10 @@ end:
     thd->set_user_connect(NULL);
   }
   mysql_mutex_unlock(&LOCK_user_conn);
+  if (error)
+  {
+    inc_host_errors(thd->main_security_ctx.get_ip()->ptr(), &errors);
+  }
   DBUG_RETURN(error);
 }
 
@@ -811,7 +735,7 @@ end:
 
   SYNOPSIS
     decrease_user_connections()
-    uc			User connection object
+    uc      User connection object
 
   NOTES
     If there is a n user connection object for a connection
@@ -869,6 +793,7 @@ void release_user_connection(THD *thd)
 }
 
 
+
 /*
   Check if maximum queries per hour limit has been reached
   returns 0 if OK.
@@ -950,7 +875,7 @@ void release_user_connection(THD *thd)
 */
 
 extern "C" uchar *get_key_conn(user_conn *buff, size_t *length,
-			      my_bool not_used __attribute__((unused)))
+            my_bool not_used __attribute__((unused)))
 {
   *length= buff->len;
   return (uchar*) buff->user;
@@ -1013,7 +938,7 @@ void reset_mqh(LEX_USER *lu, bool get_them= 0)
       USER_CONN *uc=(struct user_conn *)
         my_hash_element(&hash_user_connections, idx);
       if (get_them)
-	get_mqh(uc->user,uc->host,uc);
+  get_mqh(uc->user,uc->host,uc);
       uc->questions=0;
       uc->updates=0;
       uc->conn_per_hour=0;
@@ -1054,6 +979,14 @@ bool thd_init_client_charset(THD *thd, uint cs_number)
                      global_system_variables.character_set_client->name,
                      cs->name))
   {
+    if (!is_supported_parser_charset(
+      global_system_variables.character_set_client))
+    {
+      /* Disallow non-supported parser character sets: UCS2, UTF16, UTF32 */
+      my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), "character_set_client",
+               global_system_variables.character_set_client->csname);
+      return true;
+    }    
     thd->variables.character_set_client=
       global_system_variables.character_set_client;
     thd->variables.collation_connection=
@@ -1069,9 +1002,9 @@ bool thd_init_client_charset(THD *thd, uint cs_number)
       my_error(ER_WRONG_VALUE_FOR_VAR, MYF(0), "character_set_client",
                cs->csname);
       return true;
-    }    
+    }
     thd->variables.character_set_results=
-      thd->variables.collation_connection= 
+      thd->variables.collation_connection=
       thd->variables.character_set_client= cs;
   }
   return false;
@@ -1086,7 +1019,10 @@ bool init_new_connection_handler_thread()
 {
   pthread_detach_this_thread();
   if (my_thread_init())
+  {
+    statistic_increment(connection_errors_internal, &LOCK_status);
     return 1;
+  }
   return 0;
 }
 
@@ -1106,69 +1042,144 @@ bool init_new_connection_handler_thread()
 static int check_connection(THD *thd)
 {
   uint connect_errors= 0;
+  int auth_rc;
   NET *net= &thd->net;
 
   DBUG_PRINT("info",
              ("New connection received on %s", vio_description(net->vio)));
-#ifdef SIGNAL_WITH_VIO_CLOSE
+#ifdef SIGNAL_WITH_VIO_SHUTDOWN
   thd->set_active_vio(net->vio);
 #endif
 
-  if (!thd->main_security_ctx.host)         // If TCP/IP connection
+  if (!thd->main_security_ctx.get_host()->length())     // If TCP/IP connection
   {
+    my_bool peer_rc;
     char ip[NI_MAXHOST];
 
-    if (vio_peer_addr(net->vio, ip, &thd->peer_port, NI_MAXHOST))
-    {
-      my_error(ER_BAD_HOST_ERROR, MYF(0));
-      return 1;
-    }
-    /* BEGIN : DEBUG */
-    DBUG_EXECUTE_IF("addr_fake_ipv4",
+    peer_rc= vio_peer_addr(net->vio, ip, &thd->peer_port, NI_MAXHOST);
+
+    /*
+    ===========================================================================
+    DEBUG code only (begin)
+    Simulate various output from vio_peer_addr().
+    ===========================================================================
+    */
+
+    DBUG_EXECUTE_IF("vio_peer_addr_error",
+                    {
+                      peer_rc= 1;
+                    }
+                    );
+    DBUG_EXECUTE_IF("vio_peer_addr_fake_ipv4",
                     {
                       struct sockaddr *sa= (sockaddr *) &net->vio->remote;
                       sa->sa_family= AF_INET;
-                      struct in_addr *ip4= &((struct sockaddr_in *)sa)->sin_addr;
-                      /* See RFC 5737, 192.0.2.0/23 is reserved */
+                      struct in_addr *ip4= &((struct sockaddr_in *) sa)->sin_addr;
+                      /* See RFC 5737, 192.0.2.0/24 is reserved. */
                       const char* fake= "192.0.2.4";
                       ip4->s_addr= inet_addr(fake);
                       strcpy(ip, fake);
-                    };);
-    /* END   : DEBUG */
+                      peer_rc= 0;
+                    }
+                    );
 
-    if (!(thd->main_security_ctx.ip= my_strdup(ip,MYF(MY_WME))))
+#ifdef HAVE_IPV6
+    DBUG_EXECUTE_IF("vio_peer_addr_fake_ipv6",
+                    {
+                      struct sockaddr_in6 *sa= (sockaddr_in6 *) &net->vio->remote;
+                      sa->sin6_family= AF_INET6;
+                      struct in6_addr *ip6= & sa->sin6_addr;
+                      /* See RFC 3849, ipv6 2001:DB8::/32 is reserved. */
+                      const char* fake= "2001:db8::6:6";
+                      /* inet_pton(AF_INET6, fake, ip6); not available on Windows XP. */
+                      ip6->s6_addr[ 0] = 0x20;
+                      ip6->s6_addr[ 1] = 0x01;
+                      ip6->s6_addr[ 2] = 0x0d;
+                      ip6->s6_addr[ 3] = 0xb8;
+                      ip6->s6_addr[ 4] = 0x00;
+                      ip6->s6_addr[ 5] = 0x00;
+                      ip6->s6_addr[ 6] = 0x00;
+                      ip6->s6_addr[ 7] = 0x00;
+                      ip6->s6_addr[ 8] = 0x00;
+                      ip6->s6_addr[ 9] = 0x00;
+                      ip6->s6_addr[10] = 0x00;
+                      ip6->s6_addr[11] = 0x00;
+                      ip6->s6_addr[12] = 0x00;
+                      ip6->s6_addr[13] = 0x06;
+                      ip6->s6_addr[14] = 0x00;
+                      ip6->s6_addr[15] = 0x06;
+                      strcpy(ip, fake);
+                      peer_rc= 0;
+                    }
+                    );
+#endif /* HAVE_IPV6 */
+
+    /*
+    ===========================================================================
+    DEBUG code only (end)
+    ===========================================================================
+    */
+
+    if (peer_rc)
+    {
+      /*
+        Since we can not even get the peer IP address,
+        there is nothing to show in the host_cache,
+        so increment the global status variable for peer address errors.
+      */
+      statistic_increment(connection_errors_peer_addr, &LOCK_status);
+      my_error(ER_BAD_HOST_ERROR, MYF(0));
+      return 1;
+    }
+    thd->main_security_ctx.set_ip(my_strdup(ip, MYF(MY_WME)));
+    if (!(thd->main_security_ctx.get_ip()->length()))
+    {
+      /*
+        No error accounting per IP in host_cache,
+        this is treated as a global server OOM error.
+        TODO: remove the need for my_strdup.
+      */
+      statistic_increment(connection_errors_internal, &LOCK_status);
       return 1; /* The error is set by my_strdup(). */
-    thd->main_security_ctx.host_or_ip= thd->main_security_ctx.ip;
+    }
+    thd->main_security_ctx.host_or_ip= thd->main_security_ctx.get_ip()->ptr();
     if (!(specialflag & SPECIAL_NO_RESOLVE))
     {
-      if (ip_to_hostname(&net->vio->remote, thd->main_security_ctx.ip,
-                         &thd->main_security_ctx.host, &connect_errors))
+      int rc;
+      char *host= (char *) thd->main_security_ctx.get_host()->ptr();
+
+      rc= ip_to_hostname(&net->vio->remote,
+                         thd->main_security_ctx.get_ip()->ptr(),
+                         &host, &connect_errors);
+
+      thd->main_security_ctx.set_host(host);
+      /* Cut very long hostnames to avoid possible overflows */
+      if (thd->main_security_ctx.get_host()->length())
       {
-        my_error(ER_BAD_HOST_ERROR, MYF(0));
-        return 1;
+        if (thd->main_security_ctx.get_host()->ptr() != my_localhost)
+          thd->main_security_ctx.set_host(thd->main_security_ctx.get_host()->ptr(),
+                               min<size_t>(thd->main_security_ctx.get_host()->length(),
+                               HOSTNAME_LENGTH));
+        thd->main_security_ctx.host_or_ip=
+                        thd->main_security_ctx.get_host()->ptr();
       }
 
-      /* Cut very long hostnames to avoid possible overflows */
-      if (thd->main_security_ctx.host)
-      {
-        if (thd->main_security_ctx.host != my_localhost)
-          thd->main_security_ctx.host[min(strlen(thd->main_security_ctx.host),
-                                          HOSTNAME_LENGTH)]= 0;
-        thd->main_security_ctx.host_or_ip= thd->main_security_ctx.host;
-      }
-      if (connect_errors > max_connect_errors)
+      if (rc == RC_BLOCKED_HOST)
       {
+        /* HOST_CACHE stats updated by ip_to_hostname(). */
         my_error(ER_HOST_IS_BLOCKED, MYF(0), thd->main_security_ctx.host_or_ip);
         return 1;
       }
     }
     DBUG_PRINT("info",("Host: %s  ip: %s",
-		       (thd->main_security_ctx.host ?
-                        thd->main_security_ctx.host : "unknown host"),
-		       (thd->main_security_ctx.ip ?
-                        thd->main_security_ctx.ip : "unknown ip")));
-    if (acl_check_host(thd->main_security_ctx.host, thd->main_security_ctx.ip))
+           (thd->main_security_ctx.get_host()->length() ?
+                 thd->main_security_ctx.get_host()->ptr() : "unknown host"),
+           (thd->main_security_ctx.get_ip()->length() ?
+                 thd->main_security_ctx.get_ip()->ptr() : "unknown ip")));
+    if (acl_check_host(thd->main_security_ctx.get_host()->ptr(),
+                       thd->main_security_ctx.get_ip()->ptr()))
     {
+      /* HOST_CACHE stats updated by acl_check_host(). */
       my_error(ER_HOST_NOT_PRIVILEGED, MYF(0),
                thd->main_security_ctx.host_or_ip);
       return 1;
@@ -1176,18 +1187,43 @@ static int check_connection(THD *thd)
   }
   else /* Hostname given means that the connection was on a socket */
   {
-    DBUG_PRINT("info",("Host: %s", thd->main_security_ctx.host));
-    thd->main_security_ctx.host_or_ip= thd->main_security_ctx.host;
-    thd->main_security_ctx.ip= 0;
+    DBUG_PRINT("info",("Host: %s", thd->main_security_ctx.get_host()->ptr()));
+    thd->main_security_ctx.host_or_ip= thd->main_security_ctx.get_host()->ptr();
+    thd->main_security_ctx.set_ip("");
     /* Reset sin_addr */
-    bzero((char*) &net->vio->remote, sizeof(net->vio->remote));
+    memset(&net->vio->remote, 0, sizeof(net->vio->remote));
   }
   vio_keepalive(net->vio, TRUE);
-  
-  if (thd->packet.alloc(thd->variables.net_buffer_length))
-    return 1; /* The error is set by alloc(). */
 
-  return acl_authenticate(thd, connect_errors, 0);
+  if (thd->packet.alloc(thd->variables.net_buffer_length))
+  {
+    /*
+      Important note:
+      net_buffer_length is a SESSION variable,
+      so it may be tempting to account OOM conditions per IP in the HOST_CACHE,
+      in case some clients are more demanding than others ...
+      However, this session variable is *not* initialized with a per client
+      value during the initial connection, it is initialized from the
+      GLOBAL net_buffer_length variable from the server.
+      Hence, there is no reason to account on OOM conditions per client IP,
+      we count failures in the global server status instead.
+    */
+    statistic_increment(connection_errors_internal, &LOCK_status);
+    return 1; /* The error is set by alloc(). */
+  }
+
+  auth_rc= acl_authenticate(thd, 0);
+  if (auth_rc == 0 && connect_errors != 0)
+  {
+    /*
+      A client connection from this IP was successful,
+      after some previous failures.
+      Reset the connection error counter.
+    */
+    reset_host_connect_errors(thd->main_security_ctx.get_ip()->ptr());
+  }
+
+  return auth_rc;
 }
 
 
@@ -1217,6 +1253,7 @@ bool setup_connection_thread_globals(THD *thd)
     MYSQL_CALLBACK(thd->scheduler, end_thread, (thd, 0));
     return 1;                                   // Error
   }
+
   return 0;
 }
 
@@ -1253,10 +1290,10 @@ bool login_connection(THD *thd)
   thd->protocol->end_statement();
 
   if (error)
-  {						// Wrong permissions
+  {           // Wrong permissions
 #ifdef _WIN32
     if (vio_type(net->vio) == VIO_TYPE_NAMEDPIPE)
-      my_sleep(1000);				/* must wait after eof() */
+      my_sleep(1000);       /* must wait after eof() */
 #endif
     statistic_increment(aborted_connects,&LOCK_status);
     thd->diff_denied_connections++;
@@ -1314,7 +1351,7 @@ void end_connection(THD *thd)
 
   if (net->error && net->vio != 0)
   {
-    if (!thd->killed && thd->variables.log_warnings > 1)
+    if (!thd->killed && log_warnings > 1)
     {
       Security_context *sctx= thd->security_ctx;
 
@@ -1322,7 +1359,8 @@ void end_connection(THD *thd)
                         thd->thread_id,(thd->db ? thd->db : "unconnected"),
                         sctx->user ? sctx->user : "unauthenticated",
                         sctx->host_or_ip,
-                        (thd->stmt_da->is_error() ? thd->stmt_da->message() :
+                        (thd->get_stmt_da()->is_error() ?
+                         thd->get_stmt_da()->message() :
                          ER(ER_UNKNOWN_ERROR)));
     }
   }
@@ -1338,7 +1376,7 @@ void prepare_new_connection_state(THD* thd)
   Security_context *sctx= thd->security_ctx;
 
   if (thd->client_capabilities & CLIENT_COMPRESS)
-    thd->net.compress=1;				// Use compression
+    thd->net.compress=1;        // Use compression
 
   /*
     Much of this is duplicated in create_embedded_thd() for the
@@ -1346,7 +1384,7 @@ void prepare_new_connection_state(THD* thd)
     TODO: refactor this to avoid code duplication there
   */
   thd->proc_info= 0;
-  thd->command= COM_SLEEP;
+  thd->set_command(COM_SLEEP);
   thd->set_time();
   thd->init_for_queries();
 
@@ -1355,6 +1393,7 @@ void prepare_new_connection_state(THD* thd)
     execute_init_command(thd, &opt_init_connect, &LOCK_sys_init_connect);
     if (thd->is_error())
     {
+      Host_errors errors;
       ulong packet_length;
       NET *net= &thd->net;
 
@@ -1363,7 +1402,7 @@ void prepare_new_connection_state(THD* thd)
                         thd->db ? thd->db : "unconnected",
                         sctx->user ? sctx->user : "unauthenticated",
                         sctx->host_or_ip, "init_connect command failed");
-      sql_print_warning("%s", thd->stmt_da->message());
+      sql_print_warning("%s", thd->get_stmt_da()->message());
 
       thd->lex->current_select= 0;
       my_net_set_read_timeout(net, thd->variables.net_wait_timeout);
@@ -1384,6 +1423,8 @@ void prepare_new_connection_state(THD* thd)
       thd->server_status&= ~SERVER_STATUS_CLEAR_SET;
       thd->protocol->end_statement();
       thd->killed = THD::KILL_CONNECTION;
+      errors.m_init_connect= 1;
+      inc_host_errors(thd->main_security_ctx.get_ip()->ptr(), &errors);
       return;
     }
 
@@ -1399,7 +1440,7 @@ void prepare_new_connection_state(THD* thd)
 
   SYNOPSIS
     handle_one_connection()
-    arg		Connection object (THD)
+    arg   Connection object (THD)
 
   IMPLEMENTATION
     This function (normally) does the following:
@@ -1496,9 +1537,13 @@ void do_handle_one_connection(THD *thd_arg)
 
   for (;;)
   {
-    bool rc;
+	bool rc;
+
     bool create_user= TRUE;
 
+    NET *net= &thd->net;
+    mysql_socket_set_thread_owner(net->vio->mysql_socket);
+
     rc= thd_prepare_connection(thd);
     if (rc)
     {
@@ -1510,18 +1555,18 @@ void do_handle_one_connection(THD *thd_arg)
     {
       mysql_audit_release(thd);
       if (do_command(thd))
-	break;
+  break;
     }
     end_connection(thd);
-   
 #ifdef WITH_WSREP
-    if (WSREP(thd))
-    {
-        mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-        thd->wsrep_query_state= QUERY_EXITING;
-        mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
-    }
+  if (WSREP(thd))
+  {
+    mysql_mutex_lock(&thd->LOCK_wsrep_thd);
+    thd->wsrep_query_state= QUERY_EXITING;
+    mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
+  }
 #endif
+
 end_thread:
 #ifdef WITH_WSREP
     close_connection(thd, 0, 1);
