@@ -11,7 +11,9 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+
+#include "wsrep_sst.h"
 
 #include <mysqld.h>
 #include <sql_class.h>
@@ -20,6 +22,7 @@
 #include <sql_reload.h>
 #include <sql_parse.h>
 #include "wsrep_priv.h"
+#include "wsrep_utils.h"
 #include <cstdio>
 #include <cstdlib>
 
@@ -31,6 +34,7 @@ extern const char wsrep_defaults_file[];
 #define WSREP_SST_OPT_DATA     "--datadir"
 #define WSREP_SST_OPT_CONF     "--defaults-file"
 #define WSREP_SST_OPT_PARENT   "--parent"
+#define WSREP_SST_OPT_BINLOG   "--binlog"
 
 // mysqldump-specific options
 #define WSREP_SST_OPT_USER     "--user"
@@ -44,10 +48,12 @@ extern const char wsrep_defaults_file[];
 #define WSREP_SST_OPT_GTID     "--gtid"
 #define WSREP_SST_OPT_BYPASS   "--bypass"
 
-#define WSREP_SST_MYSQLDUMP    "mysqldump"
-#define WSREP_SST_RSYNC        "rsync"
-#define WSREP_SST_SKIP         "skip"
-#define WSREP_SST_DEFAULT      WSREP_SST_RSYNC
+#define WSREP_SST_MYSQLDUMP       "mysqldump"
+#define WSREP_SST_RSYNC           "rsync"
+#define WSREP_SST_SKIP            "skip"
+#define WSREP_SST_XTRABACKUP      "xtrabackup"
+#define WSREP_SST_XTRABACKUP_V2   "xtrabackup-v2"
+#define WSREP_SST_DEFAULT      WSREP_SST_XTRABACKUP_V2
 #define WSREP_SST_ADDRESS_AUTO "AUTO"
 #define WSREP_SST_AUTH_MASK    "********"
 
@@ -58,7 +64,6 @@ const char* wsrep_sst_donor           = "";
 
 // container for real auth string
 static const char* sst_auth_real      = NULL;
-
 my_bool wsrep_sst_donor_rejects_queries = FALSE;
 
 bool wsrep_sst_method_check (sys_var *self, THD* thd, set_var* var)
@@ -211,8 +216,8 @@ bool wsrep_sst_wait ()
 
 // Signal end of SST
 void wsrep_sst_complete (const wsrep_uuid_t* sst_uuid,
-                         wsrep_seqno_t sst_seqno,
-                         bool          needed)
+                         wsrep_seqno_t       sst_seqno,
+                         bool                needed)
 {
   if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
   if (!sst_complete)
@@ -225,18 +230,38 @@ void wsrep_sst_complete (const wsrep_uuid_t* sst_uuid,
   }
   else
   {
-    WSREP_WARN("Nobody is waiting for SST.");
+    /* This can happen when called from wsrep_synced_cb().
+       At the moment there is no way to check there
+       if main thread is still waiting for signal,
+       so wsrep_sst_complete() is called from there
+       each time wsrep_ready changes from FALSE -> TRUE.
+    */
+    WSREP_DEBUG("Nobody is waiting for SST.");
   }
   mysql_mutex_unlock (&LOCK_wsrep_sst);
 }
 
+void wsrep_sst_received (wsrep_t*            const wsrep,
+                         const wsrep_uuid_t* const uuid,
+                         wsrep_seqno_t       const seqno,
+                         const void*         const state,
+                         size_t              const state_len)
+{
+    int const rcode(seqno < 0 ? seqno : 0);
+    wsrep_gtid_t const state_id = {
+        *uuid, (rcode ? WSREP_SEQNO_UNDEFINED : seqno)
+    };
+    wsrep_init_sidno(state_id.uuid);
+    wsrep->sst_received(wsrep, &state_id, state, state_len, rcode);
+}
+
 // Let applier threads to continue
 void wsrep_sst_continue ()
 {
   if (sst_needed)
   {
     WSREP_INFO("Signalling provider to continue.");
-    wsrep->sst_received (wsrep, &local_uuid, local_seqno, NULL, 0);
+    wsrep_sst_received (wsrep, &local_uuid, local_seqno, NULL, 0);
   }
 }
 
@@ -387,6 +412,23 @@ static ssize_t sst_prepare_other (const char*  method,
   ssize_t cmd_len= 1024;
   char    cmd_str[cmd_len];
   const char* sst_dir= mysql_real_data_home;
+  const char* binlog_opt = "";
+  const char* binlog_opt_val = NULL;
+
+  if (opt_bin_log && gtid_mode > 0) {
+    binlog_opt= WSREP_SST_OPT_BINLOG;
+    if (opt_bin_logname) {
+        binlog_opt_val= (strcmp(opt_bin_logname, "0") ? opt_bin_logname : "");
+    }
+    else {
+        char buf[FN_REFLEN];
+        strmake(buf, pidfile_name, FN_REFLEN - strlen("-bin") - 1);
+        fn_format(buf, buf, "", "-bin", MYF(MY_REPLACE_EXT|MY_REPLACE_DIR));
+        binlog_opt_val=strdup(buf);
+    }
+  } else {
+      binlog_opt_val = "";
+  }
 
   int ret= snprintf (cmd_str, cmd_len,
                      "wsrep_sst_%s "
@@ -395,9 +437,15 @@ static ssize_t sst_prepare_other (const char*  method,
                      WSREP_SST_OPT_AUTH" '%s' "
                      WSREP_SST_OPT_DATA" '%s' "
                      WSREP_SST_OPT_CONF" '%s' "
-                     WSREP_SST_OPT_PARENT" '%d'",
+                     WSREP_SST_OPT_PARENT" '%d'"
+                     " %s '%s' ",
                      method, addr_in, (sst_auth_real) ? sst_auth_real : "",
-                     sst_dir, wsrep_defaults_file, (int)getpid());
+                     sst_dir, wsrep_defaults_file, (int)getpid(),
+                     binlog_opt, binlog_opt_val);
+
+  if (binlog_opt_val == NULL) {
+    free((void *)binlog_opt_val);
+  }
 
   if (ret < 0 || ret >= cmd_len)
   {
@@ -432,7 +480,6 @@ static ssize_t sst_prepare_other (const char*  method,
   return ret;
 }
 
-//extern ulong my_bind_addr;
 extern uint  mysqld_port;
 
 /*! Just tells donor where to send mysqldump */
@@ -516,7 +563,7 @@ ssize_t wsrep_sst_prepare (void** msg)
   }
   else
   {
-    ssize_t ret= guess_ip (ip_buf, ip_max);
+    ssize_t ret= wsrep_guess_ip (ip_buf, ip_max);
 
     if (ret && ret < ip_max)
     {
@@ -692,11 +739,9 @@ static int sst_donate_mysqldump (const char*         addr,
               WSREP_SST_OPT_PORT" '%s' "
               WSREP_SST_OPT_LPORT" '%u' "
               WSREP_SST_OPT_SOCKET" '%s' "
-              WSREP_SST_OPT_DATA" '%s' "
               WSREP_SST_OPT_GTID" '%s:%lld'"
               "%s",
-              user, pswd, host, port, mysqld_port, mysqld_unix_port,
-              mysql_real_data_home, uuid_str,
+              user, pswd, host, port, mysqld_port, mysqld_unix_port, uuid_str,
               (long long)seqno, bypass ? " "WSREP_SST_OPT_BYPASS : "");
 
     WSREP_DEBUG("Running: '%s'", cmd_str);
@@ -704,7 +749,9 @@ static int sst_donate_mysqldump (const char*         addr,
     ret= sst_run_shell (cmd_str, 3);
   }
 
-  wsrep->sst_sent (wsrep, uuid, ret ? ret : seqno);
+  wsrep_gtid_t const state_id = { *uuid, (ret ? WSREP_SEQNO_UNDEFINED : seqno)};
+
+  wsrep->sst_sent (wsrep, &state_id, ret);
 
   return ret;
 }
@@ -725,9 +772,9 @@ static int run_sql_command(THD *thd, const char *query)
   mysql_parse(thd, thd->query(), thd->query_length(), &ps);
   if (thd->is_error())
   {
-    int const err= thd->stmt_da->sql_errno();
+    int const err= thd->get_stmt_da()->sql_errno();
     WSREP_WARN ("error executing '%s': %d (%s)%s",
-                query, err, thd->stmt_da->message(),
+                query, err, thd->get_stmt_da()->message(),
                 err == ER_UNKNOWN_SYSTEM_VARIABLE ? 
                 ". Was mysqld built with --with-innodb-disallow-writes ?" : "");
     thd->clear_error();
@@ -750,8 +797,8 @@ static int sst_flush_tables(THD* thd)
   else
   {
     /* make sure logs are flushed after global read lock acquired */
-    err= reload_acl_and_cache(thd, REFRESH_ENGINE_LOG,
-                              (TABLE_LIST*) 0, &not_used);
+    err= reload_acl_and_cache(thd, REFRESH_ENGINE_LOG | REFRESH_BINARY_LOG,
+			      (TABLE_LIST*) 0, &not_used);
   }
 
   if (err)
@@ -894,7 +941,10 @@ wait_signal:
   }
 
   // signal to donor that SST is over
-  wsrep->sst_sent (wsrep, &ret_uuid, err ? -err : ret_seqno);
+  struct wsrep_gtid const state_id = {
+      ret_uuid, err ? WSREP_SEQNO_UNDEFINED : ret_seqno
+  };
+  wsrep->sst_sent (wsrep, &state_id, -err);
   proc.wait();
 
   return NULL;
@@ -908,6 +958,23 @@ static int sst_donate_other (const char*   method,
 {
   ssize_t cmd_len = 4096;
   char    cmd_str[cmd_len];
+  const char* binlog_opt = "";
+  const char* binlog_opt_val = NULL;
+
+  if (opt_bin_log && gtid_mode > 0) {
+    binlog_opt= WSREP_SST_OPT_BINLOG;
+    if (opt_bin_logname) {
+        binlog_opt_val= (strcmp(opt_bin_logname, "0") ? opt_bin_logname : "");
+    }
+    else {
+        char buf[FN_REFLEN];
+        strmake(buf, pidfile_name, FN_REFLEN - strlen("-bin") - 1);
+        fn_format(buf, buf, "", "-bin", MYF(MY_REPLACE_EXT|MY_REPLACE_DIR));
+        binlog_opt_val=strdup(buf);
+    }
+  } else {
+      binlog_opt_val = "";
+  }
 
   int ret= snprintf (cmd_str, cmd_len,
                      "wsrep_sst_%s "
@@ -917,13 +984,19 @@ static int sst_donate_other (const char*   method,
                      WSREP_SST_OPT_SOCKET" '%s' "
                      WSREP_SST_OPT_DATA" '%s' "
                      WSREP_SST_OPT_CONF" '%s' "
+                     " %s '%s' "
                      WSREP_SST_OPT_GTID" '%s:%lld'"
                      "%s",
                      method, addr, sst_auth_real, mysqld_unix_port,
                      mysql_real_data_home, wsrep_defaults_file,
+                     binlog_opt, binlog_opt_val,
                      uuid, (long long) seqno,
                      bypass ? " "WSREP_SST_OPT_BYPASS : "");
 
+  if (binlog_opt_val == NULL) {
+    free((void *)binlog_opt_val);
+  }
+
   if (ret < 0 || ret >= cmd_len)
   {
     WSREP_ERROR("sst_donate_other(): snprintf() failed: %d", ret);
@@ -940,7 +1013,7 @@ static int sst_donate_other (const char*   method,
   {
     WSREP_ERROR("sst_donate_other(): pthread_create() failed: %d (%s)",
                 ret, strerror(ret));
-    return -ret;
+    return ret;
   }
   mysql_cond_wait (&arg.cond, &arg.lock);
 
@@ -948,12 +1021,11 @@ static int sst_donate_other (const char*   method,
   return arg.err;
 }
 
-int wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
-                         const void* msg, size_t msg_len,
-                         const wsrep_uuid_t*     current_uuid,
-                         wsrep_seqno_t           current_seqno,
-                         const char* state, size_t state_len,
-                         bool bypass)
+wsrep_cb_status_t wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
+                                       const void* msg, size_t msg_len,
+                                       const wsrep_gtid_t* current_gtid,
+                                       const char* state, size_t state_len,
+                                       bool bypass)
 {
   /* This will be reset when sync callback is called.
    * Should we set wsrep_ready to FALSE here too? */
@@ -965,20 +1037,20 @@ int wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
   const char* data   = method + method_len + 1;
 
   char uuid_str[37];
-  wsrep_uuid_print (current_uuid, uuid_str, sizeof(uuid_str));
+  wsrep_uuid_print (&current_gtid->uuid, uuid_str, sizeof(uuid_str));
 
   int ret;
   if (!strcmp (WSREP_SST_MYSQLDUMP, method))
   {
-    ret = sst_donate_mysqldump (data, current_uuid, uuid_str, current_seqno,
-                                bypass);
+    ret = sst_donate_mysqldump(data, &current_gtid->uuid, uuid_str,
+                               current_gtid->seqno, bypass);
   }
   else
   {
-    ret = sst_donate_other (method, data, uuid_str, current_seqno, bypass);
+    ret = sst_donate_other(method, data, uuid_str, current_gtid->seqno,bypass);
   }
 
-  return (ret > 0 ? 0 : ret);
+  return (ret > 0 ? WSREP_CB_SUCCESS : WSREP_CB_FAILURE);
 }
 
 void wsrep_SE_init_grab()
@@ -988,7 +1060,10 @@ void wsrep_SE_init_grab()
 
 void wsrep_SE_init_wait()
 {
-  mysql_cond_wait (&COND_wsrep_sst_init, &LOCK_wsrep_sst_init);
+  while (SE_initialized == false)
+  {
+    mysql_cond_wait (&COND_wsrep_sst_init, &LOCK_wsrep_sst_init);
+  }
   mysql_mutex_unlock (&LOCK_wsrep_sst_init);
 }
 
