@@ -11,7 +11,9 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+
+#include "wsrep_var.h"
 
 #include <mysqld.h>
 #include <sql_class.h>
@@ -19,6 +21,7 @@
 #include <set_var.h>
 #include <sql_acl.h>
 #include "wsrep_priv.h"
+#include "wsrep_thd.h"
 #include <my_dir.h>
 #include <cstdio>
 #include <cstdlib>
@@ -34,9 +37,8 @@ const  char* wsrep_node_name        = 0;
 const  char* wsrep_node_address     = 0;
 const  char* wsrep_node_incoming_address = 0;
 const  char* wsrep_start_position   = 0;
-ulong   wsrep_OSU_method_options;
+ulong  wsrep_OSU_method_options;
 ulong   wsrep_reject_queries_options;
-static int   wsrep_thread_change    = 0;
 
 int wsrep_init_vars()
 {
@@ -59,15 +61,6 @@ bool wsrep_on_update (sys_var *self, THD* thd, enum_var_type var_type)
     // FIXME: this variable probably should be changed only per session
     thd->variables.wsrep_on = global_system_variables.wsrep_on;
   }
-  else {
-  }
-
-#ifdef REMOVED
-  if (thd->variables.wsrep_on)
-    thd->variables.option_bits |= (OPTION_BIN_LOG);
-  else
-    thd->variables.option_bits &= ~(OPTION_BIN_LOG);
-#endif
   return false;
 }
 
@@ -76,8 +69,6 @@ void wsrep_causal_reads_update (sys_var *self, THD* thd, enum_var_type var_type)
   if (var_type == OPT_GLOBAL) {
     thd->variables.wsrep_causal_reads = global_system_variables.wsrep_causal_reads;
   }
-  else {
-  }
 }
 
 static int wsrep_start_position_verify (const char* start_str)
@@ -147,7 +138,7 @@ bool wsrep_start_position_update (sys_var *self, THD* thd, enum_var_type type)
   wsrep_set_local_position (wsrep_start_position);
 
   if (wsrep) {
-    wsrep->sst_received (wsrep, &local_uuid, local_seqno, NULL, 0);
+    wsrep_sst_received (wsrep, &local_uuid, local_seqno, NULL, 0);
   }
 
   return 0;
@@ -158,7 +149,7 @@ void wsrep_start_position_init (const char* val)
   if (NULL == val || wsrep_start_position_verify (val))
   {
     WSREP_ERROR("Bad initial value for wsrep_start_position: %s", 
-		(val ? val : ""));
+                (val ? val : ""));
     return;
   }
 
@@ -174,7 +165,7 @@ static bool refresh_provider_options()
   {
     if (wsrep_provider_options) my_free((void *)wsrep_provider_options);
     wsrep_provider_options = (char*)my_memdup(opts, strlen(opts) + 1, 
-					      MYF(MY_WME));
+                                              MYF(MY_WME));
   }
   else
   {
@@ -199,7 +190,7 @@ static int wsrep_provider_verify (const char* provider_str)
     return 1;
 
   /* check that provider file exists */
-  bzero(&f_stat, sizeof(MY_STAT));
+  memset(&f_stat, 0, sizeof(MY_STAT));
   if (!my_stat(path, &f_stat, MYF(0)))
   {
     return 1;
@@ -363,7 +354,16 @@ bool wsrep_cluster_address_update (sys_var *self, THD* thd, enum_var_type type)
   bool wsrep_on_saved= thd->variables.wsrep_on;
   thd->variables.wsrep_on= false;
 
+  /* stop replication is heavy operation, and includes closing all client 
+     connections. Closing clients may need to get LOCK_global_system_variables
+     at least in MariaDB.
+
+     Note: releasing LOCK_global_system_variables may cause race condition, if 
+     there can be several concurrent clients changing wsrep_provider
+  */
+  mysql_mutex_unlock(&LOCK_global_system_variables);
   wsrep_stop_replication(thd);
+  mysql_mutex_lock(&LOCK_global_system_variables);
 
   if (wsrep_start_replication())
   {
@@ -476,7 +476,7 @@ void wsrep_node_address_init (const char* value)
 bool wsrep_slave_threads_check (sys_var *self, THD* thd, set_var* var)
 {
   mysql_mutex_lock(&LOCK_wsrep_slave_threads);
-  wsrep_thread_change = var->value->val_int() - wsrep_slave_threads;
+  wsrep_slave_count_change = var->value->val_int() - wsrep_slave_threads;
   mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
 
   return 0;
@@ -484,13 +484,10 @@ bool wsrep_slave_threads_check (sys_var *self, THD* thd, set_var* var)
 
 bool wsrep_slave_threads_update (sys_var *self, THD* thd, enum_var_type type)
 {
-  if (wsrep_thread_change > 0)
+  if (wsrep_slave_count_change > 0)
   {
-    wsrep_create_appliers(wsrep_thread_change);
-  } 
-  else if (wsrep_thread_change < 0)
-  {
-    wsrep_close_applier_threads(-wsrep_thread_change);
+    wsrep_create_appliers(wsrep_slave_count_change);
+    wsrep_slave_count_change = 0;
   }
   return false;
 }
@@ -500,11 +497,11 @@ bool wsrep_desync_check (sys_var *self, THD* thd, set_var* var)
   bool new_wsrep_desync = var->value->val_bool();
   if (wsrep_desync == new_wsrep_desync) {
     if (new_wsrep_desync) {
-      push_warning (thd, MYSQL_ERROR::WARN_LEVEL_WARN,
+      push_warning (thd, Sql_condition::WARN_LEVEL_WARN,
                    ER_WRONG_VALUE_FOR_VAR,
                    "'wsrep_desync' is already ON.");
     } else {
-      push_warning (thd, MYSQL_ERROR::WARN_LEVEL_WARN,
+      push_warning (thd, Sql_condition::WARN_LEVEL_WARN,
                    ER_WRONG_VALUE_FOR_VAR,
                    "'wsrep_desync' is already OFF.");
     }
